/***
 * Excerpted from "Modern Systems Programming with Scala Native",
 * published by The Pragmatic Bookshelf.
 * Copyrights apply to this code. It may not be used to create training material,
 * courses, books, articles, and the like. Contact us if you are in doubt.
 * We make no guarantees that this code is fit for any purpose.
 * Visit http://www.pragmaticprogrammer.com/titles/rwscala for more book information.
***/
import scala.scalanative.unsafe._
import scala.scalanative.libc._

import stdio._, string._, stdlib._
import scala.scalanative.posix.unistd._
import scala.scalanative.posix.sys.socket._
import scala.scalanative.posix.netinet.in._
import scala.scalanative.posix.arpa.inet._
import scala.collection.mutable
import scala.scalanative.posix.netdb._
import scala.scalanative.posix.netdbOps._

case class HttpRequest(
  method:String,
  uri:String,
  headers:collection.Map[String, String],
  body:String)
case class HttpResponse(
  code:Int,
  headers:collection.Map[String, String],
  body:String)

object main {
def writeRequestLine(socket_fd:Ptr[FILE], method:CString,
                     uri:CString):Unit = {
  stdio.fprintf(socket_fd, c"%s %s %s\r\n", method, uri, c"HTTP/1.1")
}

def writeHeader(socket_fd:Ptr[FILE], key:CString, value:CString):Unit = {
  stdio.fprintf(socket_fd, c"%s: %s\r\n", key, value)
}

def writeBody(socket_fd:Ptr[FILE], body:CString):Unit = {
  stdio.fputs(body, socket_fd)
}

def writeRequest(socket_fd:Ptr[FILE], request:HttpRequest):Unit = {
  Zone { implicit z =>
    writeRequestLine(socket_fd, toCString(request.method),
                     toCString(request.uri))
    for ( (key, value) <- request.headers) {
      writeHeader(socket_fd, toCString(key), toCString(value))
    }
    stdio.fputs(c"\n", socket_fd)
    writeBody(socket_fd, toCString(request.body))
  }
}

def parseStatusLine(line:CString):Int = {
  println("parsing status")
  val protocol_ptr = stackalloc[Byte](64)
  val code_ptr = stackalloc[Int]
  val desc_ptr = stackalloc[Byte](128)
  val scan_result = stdio.sscanf(line,
                             c"%s %d %s\n",
                             protocol_ptr, code_ptr, desc_ptr)
  if (scan_result < 3) {
    throw new Exception("bad status line")
  } else {
    val code = !code_ptr
    return code
  }
}

def parseHeaderLine(line:CString):(String, String) = {
  val key_buffer = stackalloc[Byte](64)
  val value_buffer = stackalloc[Byte](64)
  stdio.printf(c"about to sscanf line: '%s'\n", line)
  val scan_result = stdio.sscanf(line, c"%s %s\n", key_buffer, value_buffer)
  if (scan_result < 2) {
    throw new Exception("bad header line")
  } else {
    val key_string = fromCString(key_buffer)
    val value_string = fromCString(value_buffer)
    return (key_string, value_string)
  }
}

def readResponse(socket_fd:Ptr[FILE]):HttpResponse = {
  val line_buffer = stdlib.malloc(4096)
  println("reading status line?")
  var read_result = stdio.fgets(line_buffer, 4096, socket_fd)
  val code = parseStatusLine(line_buffer)
  var headers = mutable.Map[String, String]()
  println("reading first response header")
  read_result = stdio.fgets(line_buffer, 4096, socket_fd)
  var line_length = string.strlen(line_buffer)

  while (line_length > 2) {
    val (k,v) = parseHeaderLine(line_buffer)
    println(s"${(k,v)}")
    headers(k) = v
    println("reading header")
    read_result = stdio.fgets(line_buffer, 4096, socket_fd)
    line_length = string.strlen(line_buffer)
  }

  val content_length = if (headers.contains("Content-Length:")) {
                         println("saw content-length")
                         headers("Content-Length:").toInt
                       } else {
                         65535
                       }
  val body_buffer = stdlib.malloc(content_length + 1)
  val body_read_result = stdio.fread(body_buffer, 1, content_length,
                                     socket_fd)
  val body_length = string.strlen(body_buffer)
  if (body_length != content_length) {
    println("""Warning: saw ${body_length} bytes, but expected
                      ${content_length}""")
  }
  val body = fromCString(body_buffer)
  return HttpResponse(code, headers, body)
}



  def makeConnection(address:CString, port:CString):Int = {
    val hints = stackalloc[addrinfo]
    string.memset(hints.asInstanceOf[Ptr[Byte]], 0, sizeof[addrinfo])
    hints.ai_family = AF_UNSPEC
    hints.ai_socktype = SOCK_STREAM

    val addrInfo_ptr:Ptr[Ptr[addrinfo]] = stackalloc[Ptr[addrinfo]]
    println("about to perform lookup")
    val lookup_result = getaddrinfo(address, port, hints, addrInfo_ptr)
    println(s"lookup returned ${lookup_result}")
    if (lookup_result != 0) {
      val errString = util.gai_strerror(lookup_result)
      stdio.printf(c"errno: %d %s\n", lookup_result, errString)
      throw new Exception("no address found")
    } else {
      val addrInfo = !addrInfo_ptr
      stdio.printf(c"got addrinfo: flags %d, family %d, socktype %d, protocol %d\n",
      addrInfo.ai_family, addrInfo.ai_flags, addrInfo.ai_socktype, addrInfo.ai_protocol)

      println("creating socket")
      val sock = socket(addrInfo.ai_family, addrInfo.ai_socktype, addrInfo.ai_protocol)
      println(s"socket returned fd $sock")
      if (sock < 0) {
        throw new Exception("error in creating socket")
      }

      println("connecting")
      val connect_result = connect(sock, addrInfo.ai_addr, addrInfo.ai_addrlen)
      println(s"connect returned $connect_result")
      if (connect_result != 0) {
        val err = errno.errno
        val errString = string.strerror(err)
        stdio.printf(c"errno: %d %s\n", err, errString)

        throw new Exception("connection failed")
      }
      sock
    }
  }

  def handleConnection(sock:Int, host:String, path:String): Unit = {
    val socket_fd = util.fdopen(sock, c"r+")
    val headers = Map("Host" -> host)
    val req = HttpRequest("GET", path, headers, "")
    writeRequest(socket_fd, req)
    println("wrote request")
    fflush(socket_fd)
    val resp = readResponse(socket_fd)
    println(s"got Response: ${resp}")
    fclose(socket_fd)
  }

  def main(args: Array[String]):Unit = {
    if (args.length != 3) {
      println(s"${args.length} {args}")
      println("Usage: ./tcp_test [address] [port] [path]")
      return ()
    }

    Zone { implicit z =>
      val address = toCString(args(0))
      val host = args(0)
      val port = toCString(args(1))
      val path = args(2)
      stdio.printf(c"looking up address: %s port: %s\n", address, port)

      val sock = makeConnection(address, port)
      handleConnection(sock,host,path)
    }
  }
}

@extern
object util {
  def getaddrinfo(address:CString, port:CString, hints: Ptr[addrinfo], res: Ptr[Ptr[addrinfo]]):Int = extern
  def socket(family:Int, socktype:Int, protocol:Int):Int = extern
  def connect(sock:Int, addrInfo:Ptr[sockaddr], addrLen:CInt):Int = extern
  def gai_strerror(code:Int):CString = extern
  def fdopen(fd:Int, mode:CString):Ptr[FILE] = extern

}
